#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/lock.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/mutex.h>
#include <sys/proc.h>
#include <sys/rwlock.h>
#include <sys/uio.h>

#include <vm/vm.h>
#include <vm/pmap.h>
#include <vm/vm_extern.h>
#include <vm/vm_map.h>
#include <vm/vm_object.h>
#include <vm/vm_page.h>
#include <vm/vm_pageout.h>
#include <vm/vm_param.h>

#include "dev.h"

static d_ioctl_t dev_ioctl;
static int dev_modevent(module_t, int, void *);

static struct cdevsw dev_cdevsw = {
	.d_name = "dev",
	.d_version = D_VERSION,
	.d_flags = D_TRACKCLOSE,
	.d_ioctl = dev_ioctl,
};

static struct cdev *dev_cdev;

static int
dev_map(vm_offset_t *addr)
{

	vm_page_t m, mt;

	const size_t reserv_size = 512 * 4096;
	const size_t reserv_npages = 512;

	vm_object_t obj = vm_object_allocate(OBJT_SWAP,
	    (reserv_size * 3) >> PAGE_SHIFT);
	if (obj == NULL)
		return (ENOMEM);

	vm_offset_t vaddr;

	int error = vm_map_find(&curproc->p_vmspace->vm_map, obj, 0, &vaddr,
	    reserv_size, 0, VMFS_SUPER_SPACE, VM_PROT_ALL, VM_PROT_ALL, 0);
	if (error != KERN_SUCCESS)
		return (EFAULT);

	vm_object_reference(obj);

	VM_OBJECT_WLOCK(obj);
	vm_object_busy(obj);

	m = vm_page_alloc_contig(obj, atop(vaddr), VM_ALLOC_WIRED,
	    reserv_npages, 0, ~0, 0, 0, VM_MEMATTR_DEFAULT);
	if (m == NULL) {
		vm_object_unbusy(obj);
		VM_OBJECT_WUNLOCK(obj);

		printf("Unable to allocate superpage\n");
		return -1;
	}
	if (m->psind)
		printf("Allocated superpage\n");

	for (mt = m; mt < (m + reserv_npages); mt++) {
		vm_page_valid(mt);
		vm_page_xunbusy(mt);
	}

	error = pmap_enter(vmspace_pmap(curproc->p_vmspace), vaddr, m,
	    VM_PROT_READ | VM_PROT_WRITE,
	    VM_PROT_READ | VM_PROT_WRITE | PMAP_ENTER_WIRED, m->psind);
	vm_object_unbusy(obj);
	VM_OBJECT_WUNLOCK(obj);

	*addr = vaddr;
	return 0;
}

static int
dev_ioctl(struct cdev *dev, u_long cmd, caddr_t addr, int flags,
    struct thread *td)
{
	int error = 0;

	switch (cmd) {
	case DEVIOC_MAP:
		error = dev_map((vm_offset_t *)addr);
		break;
	default:
		error = ENOTTY;
		break;
	}

	return (error);
}

static int
dev_modevent(module_t mod, int type, void *arg)
{
	int error = 0;

	switch (type) {
	case MOD_LOAD:
		dev_cdev = make_dev(&dev_cdevsw, 0, UID_ROOT, GID_WHEEL, 0666,
		    "wiredev");
		break;
	case MOD_UNLOAD: /* FALLTHROUGH */
	case MOD_SHUTDOWN:
		destroy_dev(dev_cdev);
		break;
	default:
		error = EOPNOTSUPP;
		break;
	}

	return (error);
}

DEV_MODULE(dev, dev_modevent, NULL);
